@@ -1,5 +1,7 @@
 import re
 import os
+import math
+import csv
 import numpy as np
 import tifffile
 from datetime import datetime
@@ -252,12 +254,14 @@ def chanDic(channel):
             'RFP':'RFP',
             'CFP':'CFP',
             'DAPI':'DAPI',
+            'FR':'FR',
             'Unknown_Gray':'Unknown_Gray',
             'Unknown_Green':'Unknown_Green',
             'Unknown_Blue':'Unknown_Blue',
             'Unknown_Red':'Unknown_Red',
             'Tom_BF':'BF',
             'Tom_YFP':'YFP',
+            'Tom_GFP':'GFP',                  
             'Tom_CFP':'CFP',
             'Tom_RFP':'RFP',
             'Tom_FR':'FR',
@@ -270,14 +274,15 @@ def chanDic(channel):
             None:None
                  }
     
-    assert channel in channelDic.keys(),EM.ch1
+    assert channel in channelDic.keys(),EM.ch1 + 'Channel: '+channel
     
     return channelDic[channel]
 
 
 
 def LUTDic(channel):
-    """This takes a channel's 'general name' (see channelDic) and assigns 
+    """
+    This takes a channel's 'general name' (see channelDic) and assigns 
     an LUT mix rule (see LUTMixer()).
     You might want to change this around or add new channels etc.
     """
@@ -288,7 +293,7 @@ def LUTDic(channel):
                  'YFP':[False,True,False],
                  'CFP':[False,False,True],
                  'RFP':[True,False,False],
-                 'FR':[[True,True,False]], # = yellow!
+                 'FR':[True,True,False], # = yellow!
                  'Label':[True,True,True],
                  'Unknown_Grey':[True,True,True],
                  'Unknown_Green':[False,True,False],
@@ -304,8 +309,9 @@ def LUTDic(channel):
 
 
 def LUTMixer(mixVector):
-    """This makes an LUT in image j format from a boolean vector.
-    I.e. you give it a vector like [True, Flase, True] to say which 
+    """
+    This makes an LUT in image j format from a boolean vector.
+    I.e. you give it a vector like [True, False, True] to say which 
     RGB channels to put in the LUT.
     """
     val_range = np.arange(256, dtype=np.uint8)
@@ -336,7 +342,8 @@ def LUTInterpreter(LUT):
 
 
 def getProcessedDataDs(xPath,xSig):
-    """This function looks in the parent directory of the path given for any
+    """
+    This function looks in the parent directory of the path given for any
     directories containing the signature xSig in their name.
     It returns a set of paths to those directories.
     """
@@ -355,7 +362,8 @@ def getProcessedDataDs(xPath,xSig):
 
 
 def listStr2List(listString,convertNumeric=True):
-    """This converts a string of a python list to a list.
+    """
+    This converts a string of a python list to a list.
     Only currently works for elements that are ints or 'None'
     """
     reg = r'\[(.*)\]'
@@ -378,11 +386,12 @@ def listStr2List(listString,convertNumeric=True):
 
 
 def maskFromOutlinePath(outLinePath):
-    """ This takes the path of an image outline you have drawn and returns
-        a binary mask with all values within the outline set to 
-        1 and 0 elsewhere.
-        The only requirement is that your outline has the pixel value that 
-        is the highest in the image.
+    """ 
+    This takes the path of an image outline you have drawn and returns
+    a binary mask with all values within the outline set to 
+    1 and 0 elsewhere.
+    The only requirement is that your outline has the pixel value that 
+    is the highest in the image.
     """
     # import image
     with tifffile.TiffFile(outLinePath) as tif:
@@ -407,9 +416,10 @@ def maskFromOutlinePath(outLinePath):
 
 
 def shapeFromFluoviewMeta(meta):
-    """ This gets the 7D dimensions from fluoview metadata. 
-        In some fluoview versions it doesn't include the dimension 
-        name if the dimension size is 1 so we have to add it.
+    """ 
+    This gets the 7D dimensions from fluoview metadata. 
+    In some fluoview versions it doesn't include the dimension 
+    name if the dimension size is 1 so we have to add it.
     """
     dims = meta['Dimensions']
     dimsDic = {l[0]:l[1] for l in dims}
@@ -458,8 +468,9 @@ def tif2Dims(tif):
 
 
 def meta2StartMom(meta):
-    """ This takes a session's metadata file and returns a datetime object 
-        of the moment when the file was started.
+    """ 
+    This takes a session's metadata file and returns a datetime object 
+    of the moment when the file was started.
     """
     # the format of the datatime string we give it
     TX = '%d/%m/%Y %H:%M:%S'
@@ -491,8 +502,9 @@ def meta2StartMom(meta):
 
 
 def meta2TStep(meta):
-    """ This takes a session's metadata file and returns a timedelta object
-        of the time between time points.
+    """ 
+    This takes a session's metadata file and returns a timedelta object
+    of the time between time points.
     """
     # time interval group(1), units group(2)
     DTReg = re.compile(r'Repeat T - \d+ times? \((\d+) (\w+)\)')
@@ -515,7 +527,8 @@ def meta2TStep(meta):
 
 
 def onlyKeepChanges(theList):
-    """ This makes a list from theList in which only elements which are 
+    """ 
+    This makes a list from theList in which only elements which are 
     different from the previous are kept.
     """
     
@@ -533,7 +546,8 @@ def onlyKeepChanges(theList):
 
 
 def savedByXFoldQ(filepath):
-    """ Returns True if the files was saved by this package, False otherwise
+    """ 
+    Returns True if the files was saved by this package, False otherwise
     Tests this by looking at metadata.
     """
     with tifffile.TiffFile(filepath) as tif:
@@ -549,48 +563,108 @@ def savedByXFoldQ(filepath):
 def saveTiffForIJ(
     outPath,
     data,
-    chan,
-    seshDs,
+    chan=False,
+    seshDs=False,
     autoscale=True,
     minP=2,
     maxP=98,
     overwrite=False):
     """
     Uses tifffile package to save 7D data in way that is good for imagej.
-    Also gives metadata that xfold will recognise.
+    Also saves metadata inside the tif that xfold will recognise.
     
-    seshDs should be [seshT,seshF,seshM,seshZ,seshC]
-    
-    autoscale here doesn't change the data, just saves some metadata that 
-    imagej reads and sets the scale with.
+    Parameters
+    ----------
+    outPath : str
+        File path to save to.
+    data : numpy.array
+        Image array. Must have 7 dimensions as in standard TData structure.
+    chan : list of str
+        The channel names. Used to set LUT. If not provide it will try to set 
+        default channels but currently this only works if the number of 
+        channels can be read from provided seshDs or the data shape. It only 
+        knows how to read from the data shape if the shape has 2,3 or 7 axes.
+    seshDs : list of lists
+        Used to save as metadata, should be: [seshT,seshF,seshM,seshZ,seshC]. 
+        Also used to try and reshape data to 7D if it is not 7D and set 
+        default channels if they're not provided. If seshDs not provided then 
+        it can continue only if the data shape has 2,3 or 7 axes.
+    autoscale : bool
+        Whether to autoscale or not. This doesn't change the data, it just 
+        sets the metadata for imagej.
+    minP : int [0-100]
+        The percentile to set the min pixel value to during autoscaling.
+    maxP : int [0-100]
+        The percentile to set the max pixel value to during autoscaling.
+    overwrite : bool
+        Whether to allow overwriting of data with the same filename.
+        
+    Notes
+    -----
+    It will try to reshape the data if it isn't 7D - it will work if 
+    images are properly ordered and you give the correct seshDs. It will also 
+    work if the data is 3D - it will assume the first axis is channels.
     """
-    dims = data.shape
-    assert len(dims)==7,'error from saveTiffForIJ: data must be 7D'
     
-    # if this path already exists we better have set overWrite to True
-
+    dims = data.shape
+    assert len(dims)<8 and len(dims)>1,'data must be maximum 7D and minimum 2D'
+    
+    defaultChans = ['BF','GFP','RFP','CFF','FR']
+    if not chan:
+        if seshDs:
+            assert len(seshDs[4])<6,EM.sv4
+            chan = defaultChans[:len(seshDs[4])]
+        elif len(dims)==7:
+            assert dims[-3]<6,EM.sv5
+            chan = defaultChans[:dims[-3]]
+        elif len(dims)==3:
+            assert dims[0]<6,EM.sv5
+            chan = defaultChans[:dims[0]]
+        elif len(dims)==2:
+            chan = ['BF']          
+        else:
+            raise Exception(EM.sv3)
+    
+    # reshape data according to seshDs if data is not 7D and seshDs provided
+    if len(dims)!=7 and seshDs:
+        seshDL = [len(s) for s in seshDs]
+        nImMeta = np.product(seshDL)*dims[-1]*dims[-2]
+        assert np.product(dims)==nImMeta,EM.sv2
+        dims = seshDL + [dims[-2]] + [dims[-1]]
+        data = data.reshape(dims)
+    # reshape data if data has 3 axes and seshD not provided 
+    # also set seshDs from dims
+    elif len(dims)==3 and not seshDs:
+        dims = [1,1,1,1]+[dims[-3]]+[dims[-2]]+[dims[-1]]
+        data = data.reshape(dims)
+        seshDs = [['None']*dims[0],
+                  ['None']*dims[1],
+                  ['None']*dims[2],
+                  ['None']*dims[3],
+                  ['None']*dims[4]]  
+    # reshape data if data has 2 axes and seshD not provide    
+    # also set seshDs from dims
+    elif len(dims)==2 and not seshDs:
+        dims = [1,1,1,1,1]+[dims[-2]]+[dims[-1]]
+        data = data.reshape(dims)
+        seshDs = [['None']*dims[0],
+                  ['None']*dims[1],
+                  ['None']*dims[2],
+                  ['None']*dims[3],
+                  ['None']*dims[4]]          
+    # set seshDs if data has 7 axes and seshDs not provided
+    elif len(dims)==7 and not seshDs:
+        seshDs = [['None']*dims[0],
+                  ['None']*dims[1],
+                  ['None']*dims[2],
+                  ['None']*dims[3],
+                  ['None']*dims[4]] 
+    elif len(dims)!=7 and not seshDs:
+        raise Exception(EM.sv6)
+        
     assert not (os.path.exists(outPath) and not overwrite),EM.sv1
     
     # this is the metadata to add:
-    # you don't add channels, slices or frames b/c it does 
-    # it automatically from the array shape!
-    # this stuff goes into what imagej calls 'Image Description:', it
-    # reads this when opening, they control display parameters
-    # if you want to find what words to use to save a new parameter
-    # change it in imagej and save and reopen to check imagej 
-    # really does remember that stuff
-    # then turn on imagej debug mode (Edit -> Options -> Misc)...
-    # ... and open that image again: the parameter you changed should
-    # appear in Image Description somewhere
-    # this doesn't get everything! some things are stored in binary 
-    #(see ijmeta below)... 
-    # these are seen in debug mode as memory locations
-    # e.g. to get the display ranges for each channel I: 
-    # changed file in imagej, saved new file and opened in 
-    # python with tifffile, imagej_metadata 
-    # this is also where we write our own metadata
-    # we need to do things slightly differntly if it is a single 
-    # image vs a hyperstack
     singleImQ = all([d==1 for d in dims[0:5]])
     meta = {'hyperstack':'true','mode':'composite','unit':'um',
             'spacing':'2.56','loop':'false','min':'0.0','max':'256',
@@ -643,7 +717,7 @@ def getTagDic(fp,madeBy):
     the values are the tags. 
     
     Our code in general is going to assume that all files with the same tag 
-    contain equivalent Q-points. That is, all files tih time tag '_t0004' 
+    contain equivalent Q-points. That is, all files with time tag '_t0004' 
     have the same set of time points even if they contain different subsets 
     of other dimensions, e.g. you could have _t0004_m0000 and _t0004_m0001 
     but both contain time points e.g. t=6,7,8.
@@ -731,41 +805,298 @@ def UDRL2LRUD(I,NY,NX):
     return [(NX-1-c)*NY + r for i,r,c in zip(I,R,C)]
 
 
-def removePads(im,retPad=False):
+def removePads(im,pads=None,retPad=False):
     """
-    This detects and removes padding from an image. Padding is defined as 
+    Removes padding from a 2D image by either detecting the blank (0) edges 
+    or according to user provided padding. Padding is defined as 
     rows/columns of 0. It returns the image without padding and (optionally) 
     the padding that was removed so you can add it back later if you like.
     
     Parameters
     ----------
-    im : numpy.array
-    
+    im : 2D numpy.array
+        The image.
+    pads : None or [int,int]
+        If None then it will find the padding by detecting blank rows/columns 
+        at edges. Otherwise it gives the total number of blank pixels that are 
+        to be removed from the [y,x]-axes. See code for how this total number 
+        is divided between the two sides.
     retPad : bool
         Whether to return the pad along with cropped image.
     
     Returns
     -------
-    im : numpy.array
+    im : 2D numpy.array
         The image without padding.
     pad : tuple of tuple of ints
         If returnPad then how many pixels were removed is returned in: 
         (top,bottom,left,right). In this case im and pad are a tuple.
     """
     ny,nx =im.shape
-    for t in range(ny):
-        if any(im[t,:]):
-            break
-    for b in reversed(range(ny)):
-        if any(im[b,:]):
-            break
-    for l in range(nx):
-        if any(im[:,l]):
-            break
-    for r in reversed(range(nx)):
-        if any(im[:,r]):
-            break
+    if pads is None:
+        for t in range(ny):
+            if any(im[t,:]):
+                break
+        for b in reversed(range(ny)):
+            if any(im[b,:]):
+                break
+        for l in range(nx):
+            if any(im[:,l]):
+                break
+        for r in reversed(range(nx)):
+            if any(im[:,r]):
+                break
+    else:
+        t,b = [pads[0]//2,ny-math.ceil(pads[0]/2)-1]
+        l,r = [pads[1]//2,nx-math.ceil(pads[1]/2)-1]
     if retPad:
         return (im[t:b+1,l:r+1],(t,ny-b-1,l,nx-r-1))
     else:
-        return im
\ No newline at end of file
+        return im[t:b+1,l:r+1]
+    
+    
+def pad2Size(im,sizes):
+    """
+    Pads the image to the specified size, using the convention of 
+    sessionStitcher to decide where to put extra pixels.
+    
+    Parameters
+    ----------
+    im : array-like
+        The image you want to pad. Can be any dimension, always the last two 
+        axes are the ones that will be padded.
+    sizes : tuple
+        The (ysize,xsize) that you pad to.
+    
+    Returns
+    -------
+    im : array like
+        The padded image.
+    """
+    pY = sizes[0] - im.shape[-2]
+    pX = sizes[1] - im.shape[-1]
+    pY = ((pY//2,math.ceil(pY/2)),)
+    pX = ((pX//2,math.ceil(pX/2)),)
+    pad = ((0,0),)*(im.ndim - len(sizes)) + pY + pX
+    return np.pad(im,pad)  
+
+
+def takeWindow(im,win):
+    """
+    Extracts window from an image when window is given in format of 
+    XFold.WindowDic, which is [[y,x],[y,x],[y,x],[y,x]]. Image is assumed to 
+    be size of the template that the window was drawn on, no padding 
+    considered.
+    
+    Parameters
+    ----------
+    im : numpy.array
+        2D image that will be cropped.
+    win : list of lists of ints
+        The window in XFold.WindowDic format.
+    """
+    return im[win[0][0]:win[2][0],win[0][1]:win[2][1]]
+
+
+def makeuint8(img,quantH=0.9,quantL=False):
+    """
+    Converts uint16 to uint8 with re-scaling to the max value. Can also do 
+    clipping (max or max and min) to stop detail loss from anomalous high 
+    values.
+    """
+    maxV = np.quantile(img,quantH)
+    img[img>maxV] = maxV
+    if quantL:
+        minV = np.quantile(img,quantL)
+        img[img<minV] = minV
+    if quantL:
+        return (255*((img-minV)/(maxV-minV))).astype('uint8')
+    else:
+        return (255*(img/maxV)).astype('uint8')
+    
+    
+def ExtractProfileRectangle(data,outDir,outName=False,masks=None,
+                            NXSegs=10,downsizeY=2,overwrite=False):
+    """
+    This saves csvs containing averaged pixel values taken from the image 
+    data. NXSegs and downsizeY define a grid in which the mean of each 
+    square is returned.
+
+    Parameters
+    ----------
+    data : array-like, shape (NIm,sizeY,sizeX)
+        The image data. One csv will be saved for each image along axis 0.
+    outDir : str
+        Path of directory to save csvs too. This directory will be created if 
+        needed.
+    outName : str or list of str
+        The name of the csv(s). If False then it will create a name and number 
+        according NIm. If you provide a list then the names correspond to the 
+        images along axis 0 of the data.
+    masks : array-like, shape (NIm,sizeY,sizeX) or (sizeY,sizeX)
+        Image(s) which define(s) a region of arbitrary shape to decide which 
+        pixels to include. This image should be binary with included regions 
+        as white.
+    NXSegs : int
+        The number of segments to return along the x-axis. All pixel
+        values are averaged within each segment. The X position in pixels
+        of the middle of each segment will be given in the CSV row heading
+        (measured relative to the window).
+    downsizeY : int
+        Factor to downsize y-axis length by. Pixels are averaged. Position
+        put in CSV headings too.
+    overwrite : bool
+        Whether to allow csvs to be overwritten.
+
+    Returns
+    -------
+    Nothing returned but csvs are saved. The csvs have NSegs+1 columns and
+    1+NY//downsizeY rows. One separate csv for each image along data axis 0.
+    """
+    if len(data.shape)==2:
+        data = data[np.newaxis]
+    NIm,sizeY,sizeX = data.shape
+    for i in range(NIm):
+        # initiate np array and steps for collecting data
+        dy = downsizeY
+        NY = int(sizeY/dy)
+        dx = int(sizeX/NXSegs)
+        _csv = np.zeros((NY,NXSegs))
+        for y in range(NY):
+            for x in range(NXSegs):
+                # input the data
+                blockD = data[i,y*dy:(y+1)*dy,x*dx:(x+1)*dx].copy()
+                if len(masks.shape)==3:
+                    blockM = masks[i,y*dy:(y+1)*dy,x*dx:(x+1)*dx].copy()
+                else:
+                    assert len(masks.shape)==2,'masks not of shape we can handle'
+                    blockM = masks[y*dy:(y+1)*dy,x*dx:(x+1)*dx].copy()
+                result = blockD[blockM==1]
+                if result.size==0:
+                    _csv[y,x] = 'NaN'
+                else:
+                    _csv[y,x] = np.mean(result)
+                del blockD
+                del blockM
+        del data
+    
+        # add x-slice and y-distance headings
+        xlab = [str(dx*x+dx//2) for x in range(NXSegs)]
+        xlab = np.array(['x='+x+' pixels' for x in xlab])
+        ylab = ['y='+str(dy*y+dy//2)+'pixels' for y in range(NY)]
+        ylab = np.array(['Y distance']+ylab)
+        _csv = np.vstack((xlab,_csv))
+        _csv = np.hstack((ylab.reshape((NY+1,1)),_csv))
+    
+        if not outName:
+            outName1 = 'ExtractedValues_Image_'+str(i).zfill(4)+'.csv'
+        else:
+            if isinstance(outName,list):
+                assert len(outName)==NIm,EM.ae9
+                outName1 = outName[i]
+            else:
+                outName1 = outName
+        if not os.path.exists(outDir):
+            os.makedirs(outDir)
+        outPath = os.path.join(outDir,outName1)
+        if not overwrite:
+            assert not os.path.exists(outPath),EM.ae5
+    
+        with open(outPath,'w',newline='') as file:
+            writer = csv.writer(file)
+            writer.writerows(_csv)
+
+            
+            
+def adjustFilterEdges(filt,DF,tileA,tileB,molap):
+    """
+    This improves the homogenisation filter using the knowledge that the 
+    bottom fraction of tileA overlaps with the top section of tileB, with a 
+    fractional overlap size of molap.
+    
+    Parameters
+    ----------
+    filt : array-like shape (NY,NX)
+        The original filter, e.g. calculated by BaSiC.
+    DF : array-like shape (NY,NX)
+        The original dar-fileld filter, e.g. calculated by BaSiC.        
+    tileA : array-like shape (NY,NX)
+        A tile in the montage where there is good signal all at the bottom.
+    tileB : array-like shape (NY,NX)
+        The tile in the montage directly below tileA.
+    molap : float
+        The decimal fraction of tile A and tileB that overlap.
+    """
+    
+    # extract section of tiles and filter that should overlap
+    ny,nx = tileA.shape
+    segA_B = (tileA-DF)[ny-math.floor(ny*molap)-1:-1]
+    segB_T = (tileB-DF)[:math.floor(ny*molap)]
+    segF_B = filt[ny-math.floor(ny*molap)-1:-1]
+    segF_T = filt[:math.floor(ny*molap)]
+    
+    ny_seg = segA_B.shape[0]
+    
+    # y is essentially the equation that says that the corrected image sections 
+    # should be the same... solve that equation with best fit straight line
+    # that gives you the factor to correct the filter with
+    x = np.array(range(ny_seg))
+    y = np.mean((segA_B/segB_T)/(segF_B/segF_T),axis=1)
+    a,b = np.polyfit(x,y,1)
+    
+    A = np.ones((ny))
+    for i in range(ny):
+        if i > ny - ny_seg:
+            A[i] = math.sqrt(a*(i - (ny - ny_seg)) + b)
+        elif i < ny_seg:
+            A[i] = 1/math.sqrt(a*i + b)            
+    
+    start_T = int(3*ny_seg/4)
+    end_T = int(5*ny_seg/4)
+    max_T = A[start_T]
+    for i in range(start_T,end_T):
+        A[i] = max_T + (i-start_T)*((1-max_T)/(ny_seg/2))
+        
+    start_B = int(ny-ny_seg-(ny_seg/4))
+    end_B = int(ny-ny_seg+(ny_seg/4))
+    max_B = A[end_B]
+    for i in range(start_B,end_B):
+        A[i] = 1 + (i-start_B)*((max_B-1)/(ny_seg/2))        
+        
+    A = A[:,np.newaxis]
+    A = np.tile(A,(1,nx))
+    
+    filt = filt*A
+    filt = filt/np.mean(filt)
+    
+    return (A,filt)
+
+
+
+def signalNoiseMasks(mask,mask2=None):
+    """
+    You provide a mask and it returns two masks, one of definite signal 
+    and one of definite noise. I.e. it erodes/dilates/inverts the provided 
+    mask of signal to be sure.
+    
+    mask : array-like
+        The mask of a segmentation of signal.
+    mask2 : False or array-like or str
+        A further mask to restrict the region from which signal is taken.
+    
+    Returns
+    -------
+    sig : array-like
+        The mask of where signal definitely is.
+    noise : array-like
+        The mask of where noise definitely is.
+    """
+    # nuclei and noise masks
+    kernel = np.ones((3,3),np.uint8)
+    sig = cv.erode(mask,kernel)
+    mask2 = cv.erode(mask.astype('uint16'),kernel)
+    nucleiMask = np.logical_and(nucleiMask,mask2)
+    noise = 1-cv.dilate(mask,kernel)
+    noise = np.logical_and(noise,mask2)
+    
+    return sig,noise
\ No newline at end of file
